// includes defines, globals

#include "C:\Program Files\PICCa\Projects\lf2523\main.h"

// don't define this to just use the master oscillator
// do define it if you've hooked up a crystal to timer1 pins
#DEFINE WATCHXTAL // comment out to use internal main clock for seconds timing

#IFDEF WATCHXTAL
// order to check on interrupt flags, gives the internal dispatcher a hint
#FUSES LPT1OSC   //Timer1 configured for low-power osc operation
#PRIORITY INT_AD, INT_TIMER3, INT_TIMER1,INT_TBE
#ELSE // no watch crystal so fuses, etc, different
#FUSES NOLPT1OSC // don't low-power timer1 so it can go faster in this case
#PRIORITY INT_AD, INT_TIMER3, INT_TBE
#ENDIF

// Data EEPROM -- stuff defaults in here

#ROM 0xF00000={"P I am here!\0"}

// globals in ram

////////////// time stuff

int1 secflag; // set by ISR on the second, cleared by main when read
int8 seconds;
int8 minutes;
int16 hours; // 7.481278539 years worth
int8 even; // temp for toggles
#define NUM_ONESHOTS 8
int8 oneshots[NUM_ONESHOTS];

////////////////////////// assorted other globals
int8 i,j; // used all over main, loop index
int8 whyboot; // capture reason for reset into this
char tmpstr[32];

#DEFINE AD_CH 4

int16 pwmduty = 128;
int16 adres[AD_CH]; // size = number of a/d channels
int16 adresb[AD_CH];
int1 badsweep; // a/d busy flag

//char tmpstr[32]; // for strings, test buffer
//char *strptr;
///////////////////////// isr stuff -- warning, different thread(s)

// our version of "hungrarian notation" is to preface with i on isr variables.
int8 ii; // used in ISR threads, loop index
int8 iadidx;

// serial port stuff
#DEFINE SERIALSIZE 64 //-- big tradeoff here, uses lots of ram
char iPCXmit[SERIALSIZE]; // serial port transmit buffer
int8 iPCXmiti; // current index into buffer
int1 iPCXmitbsy; // transmit-busy flag
char PCXmitbuf[SERIALSIZE]; // user buffer to sprintf to

volatile char iPCrcv[SERIALSIZE];  // place for rcv isr to stuff chars
char PCrcv[SERIALSIZE]; // isr copies to this for background uses
int8 ircvindex;         // rcv buffer index
volatile int1 ircvflag;          // set when data available to parse

int8 PCState;     // state variable for PC parsing machine


#IFNDEF WATCHXTAL
int8 isubseconds; // for div by 100 in timer isr (get rid of 32k xtal)
#ENDIF



//********************************//////// subroutines
void bufferAD()
{
// grabs most current a/d sweep to a safe buffer for background
 while(badsweep)restart_wdt(); // wait for a full sweep
 DISABLE_INTERRUPTS(INT_AD);
 for (i=0;i<AD_CH;i++) adresb[i]= adres[i];
 ENABLE_INTERRUPTS(INT_AD);
}

//////////////////////////
void ZeroTime()
{
#IFNDEF WATCHXTAL // in other words, if no xtal on timer 1
 isubseconds = 0;
#ENDIF 

 secflag = FALSE;
 badsweep = FALSE;
 seconds = minutes = 0;
 hours = 0;
 for (i=0;i<NUM_ONESHOTS;i++) oneshots[i] = 0; // clear out timers
 for (i=0;i<AD_CH;i++) adres[i] = 0; // clear out a/d results

 even = 0; // toggle flag
 iadidx = 0; // current a/d channel for aq
}

// init uses this, so it wants to be first
/**
This is how you send ascii to the serial port, don't use this
if you're using printf() or vice versa.  This one is non blocking
if not already busy, and interrupt driven so it won't take
polling cycles away from the main loop code, so it's preferred.

However, sprintf(), used with this, itself can take some time,
the ascii conversions can be code intensive.

It's on you to keep your strings inside the current buffer size.
For speed, no checking is done (so don't overwrite other memory either!)

Note, for the first few millisconds after boot, the baud rate
will be wrong, as the internal clock/pll gets stable.
*/
void SendPC (char* yourbuf) // call with string buffer pointer, nul terminate
{ // don't call if we are busy, and don't use printf if using this
 while(TRUE == iPCXmitbsy); // you're supposed to check this before calling me!
 // (or just know UART can't still be busy from last time)
 //above should cause a WDT timeout to warn ignorant programmers!
 strcpy(iPCXmit,yourbuf); // copy your string to our isr buffer
 bit_set (TXSTA,5); // turn on port enable (TXEN)
 TXREG = iPCXmit[0]; // send first character
 iPCXmiti = 1;  // because we just sent the first one, isr does the rest
 iPCXmitbsy = TRUE; // we're busy now, leave us alone
 bit_set(PIE1,4); // set interrupt enable for transmit (TXIE)
 // some PICS will take 2-3 chars right off, if that's true here, we'll
 // get interrupt(s) right away to fill the hardware up.
 // yourbuf is now ready to use again, regardless.
}

//////////////////////////
/* 

Set hardware and memory up here.  This will have to change when you use
a different PIC, of course, to match the hardware you've got and what's
hooked to what.

The hardware guy should fix things so the way the pic powers up doesn't do
bad things, and that's easy -- pretty much everything comes up float/input,
so you can have pullups-pulldowns as required for this.

The software guy can set things up as desired before changing any i/o pins
to outputs.  No smoke need be released from anything, ever.

Not set up now to use priority interrupts (which can interrupt low pri ones)
IF that is changed, make sure no interruptable interrupt shares variables with
a high priority one. For now, we do that to save ram.  Beware.

Not using "fast" interrupts either for similar reasons.  Use this 
as is till something better is really needed, it's pretty quick.

*/
void init()
{
   setup_oscillator(OSC_32MHZ,0);  // try internal RC with pll
// Get rid of the xtal -- will it be accurate?
// Well at any rate, it does work, and speed seems good
   whyboot = restart_cause();  // capture info on why we are here
   ZeroTime();  // runtime starts now -- all time stuff cleared, mem setup

// we use fast_io and do these things ourself
   set_tris_a(0b00001111);  // 4 a/d inputs, clk? now internal RC
   port_b_pullups(TRUE);  // keeps down noise on inputs
   set_tris_b(0b11111111);
   set_tris_c(0b11010010); // rs 232, oscillator, pwm/ccp1
   
   setup_adc_ports(AN0_TO_AN3 | VSS_VDD);  // 4 channels
   setup_adc(ADC_CLOCK_DIV_32 | ADC_TAD_MUL_4); // as fast as possible?
   set_adc_channel(iadidx); // start on chnl zero
   enable_interrupts(INT_AD); // a/d interrupts do the channel sweep
//   setup_spi(SPI_SS_DISABLED); // temp
   setup_wdt(WDT_OFF);        // temp  
   setup_timer_0(RTCC_OFF);   // temp

#IFDEF WATCHXTAL // use timer 1 pins for watch crystal oscillator
// @@@ this code needed if using external 32768hz xtal, otherwise not
   setup_timer_1(T1_EXTERNAL_RTC); // oscillate with 32768 hz clock xtal
   enable_interrupts(INT_TIMER1); // enables the timer1 interrupt
// end optional external xtal code
#ENDIF // else do whatever you want with timer 1

   setup_timer_3(T3_INTERNAL | T3_DIV_BY_8); // counts at 1 mhz
   enable_interrupts(INT_TIMER3); // allow the interrupt
   
   setup_timer_2(T2_DIV_BY_1,0x3f,1); // ~123 khz for pwm, 8 bit rez
   setup_ccp1(CCP_PWM); // start pwm output
   set_pwm1_duty (pwmduty); // start with sq wave (diagnostic)
 
   setup_comparator(NC_NC_NC_NC); // temp
   setup_vref(FALSE); // temp
   
   iPCXmitbsy = FALSE; // make sure serial transmit busy flag is false
   iPCXmit[0] = 0;

   ircvindex = 0; // set up rcv isr variables
   ircvflag = FALSE; // can't be true yet, we haven't set up the port
   PCState = 0; // no messages yet
   
   bit_set (RCSTA,4);  // continuous recieve enable
   bit_set (RCSTA,7); // port enable (may already be set by #use RS232)
   bit_set (PIE1,5);  // interrupt on receive enable RCIE

   delay_ms(255);  // wait awhile for things to get stable.   

   enable_interrupts(GLOBAL); // start them now, see below
   strcpy(PCXmitbuf,"\r\nC-Lab opsys 10.11 Restart cause: ");  // build startup msg
   switch (whyboot)
   { // tell user why we are coming up (again?)
   case NORMAL_POWER_UP:    
      strcpy(tmpstr,"Normal power up.");
      break;
   case WDT_TIMEOUT:   
      strcpy(tmpstr,"WDT_TIMEOUT.");
      break;
   case MCLR_FROM_SLEEP:    
      strcpy(tmpstr,"MCLR_FROM_SLEEP");
      break;
   case MCLR_FROM_RUN:    
      strcpy(tmpstr,"MCLR_FROM_RUN.");
      break;
   case BROWNOUT_RESTART:   
      strcpy(tmpstr,"BROWNOUT_RESTART.");
      break;
   case WDT_FROM_SLEEP:     
      strcpy(tmpstr,"WDT_FROM_SLEEP.");
      break;
   case RESET_INSTRUCTION:
      strcpy(tmpstr,"RESET_INSTRUCTION.");
      break;   
   }// end switch
   strcat(PCXmitbuf,tmpstr);
   SendPC(PCXmitbuf);     
}

//////////////////////////////////********** interrupts *******************

#int_RDA // serial port recieve data, ascii, CR or LF terminated
void Rcv232()
{
 ii = RCREG; // get the data byte that came in
 if ('\r' != ii && '\n' != ii && 0 != ii)
 { // it's a legit character in body of message
   iPCrcv[ircvindex++] = ii; // so store it in the buffer 
 } else // some crap linefeed, cr or null, end of a message if there was one
 {
  if (0!= ircvindex) // was there a message at all, or was this a second ending
  { // we ignore extra message-end characters and nulls
   iPCrcv[ircvindex] = 0; // add null termination to the string
   strcpy (PCrcv,iPCrcv); // copy to background buffer (@@@ sort of safe!)
   ircvindex = 0;  // ready for next message
   ircvflag = TRUE; // tell background to grab this before the next wipes it out
  } // we just do nothing for a second line terminator or null...ignore it
 }

}
////////////////////////////////////
#IFDEF WATCHXTAL // uses timer one for watch xtal osc and runtime
// this code needed with external xtal on timer 1, else not
#int_timer1 // flag each passing second
void timer_1_interrupt()
{ // important to reload timer in a timely way
 bit_set(TMR1H,7);   // makes a bsf, but no TMR1L access, and no effect
 secflag = TRUE;     // set flag for background, let it do the calx for HMS,
 //can then read them without ever having a carry error during capture
}
// end xtal runtime code
#ENDIF // else, use timer1 for whatever you want to

//////////////////////////////////**********
#int_timer3 // used for faster timings, 10 ms oneshots
void timer_3_interrupt()
{ // should get here at 100 hz more or less, reload timer, do oneshots
set_timer3(55603); // set to count 10k times per interrupt
// tune above number for more accurate time, it was off ~4% with the "correct"
// number of 55535 -- this has better rez than the osctune function

badsweep = TRUE; // flag for background, we're busy now
read_adc(ADC_START_ONLY); // setup elesewhere, start a  channel sweep here
// sweep is completed and the flag cleared in the a/d isr.

#IFNDEF WATCHXTAL
//@@@ this code generatees runtime off the internal RC oscillator to save parts
if ( 0 == isubseconds)
{
 isubseconds = 100; //  time for one second, more or less (exact we hope)
 secflag = TRUE;
} 
isubseconds--; // count down for next seconds flag
// if using external xtal on timer1, don't use this, see above
#ENDIF

/* This is real important code below.  Apps use these by setting a oneshot to
some number, then when they are called, if it isn't zero yet, they can yield
to the opsys right away, instead of spin-looping and wasting cycles when some
other app might need them.  So in a way, this one line of code is what makes
it all work!  That and something like "if (0 != onshot[mine]) return;" in the 
app code.  Don't be fooled that it's only one line, it's the lines we don't need
that define how slick the design is!
*/

for (ii=0;ii<NUM_ONESHOTS;ii++) 
   if (oneshots[ii]) oneshots[ii]--; // count down if not zero
// end crucial code that does most of the timing work for everyone.

// diagnostic junk, blink a red-green led connected between two port pins
even++;
if (even & 1) // setup a toggle, diagnostic code
   {
    OUTPUT_HIGH(PIN_C3);// green led
    OUTPUT_LOW(PIN_C5);
   } else
   {
    OUTPUT_HIGH(PIN_C5); // red led
    OUTPUT_LOW(PIN_C3);
   }

}// end timer3 int
///////////////////////////////
#int_AD
void AD_interrupt() // gather a/d channels into an array for bckgnd
{ // sweep triggered by timer at 100hz for now
 adres[iadidx++] = read_adc(ADC_READ_ONLY) >> 4; // get a/d channel into array
 // shift to right-justify, 0-4095 range results
 iadidx &= (AD_CH-1); // limit it to the channels we've set up  @@@ #defines?
 set_adc_channel(iadidx); // point to next channel to acquire
 if (0!=iadidx) read_adc(ADC_START_ONLY); // get next chnl in sweep if not done
 else badsweep = FALSE; // we're done, so let background know it's safe to read.
}

///////////////////////////////
#INT_TBE // serial port transmit hardware buffer empty, see if there's more
void XmitInterrupt()
{ // send next character in buffer if not null, or clear busy flag when done
ii = iPCXmit[iPCXmiti]; // get next char from buffer
if (0 != ii) // check for end of string
{
 TXREG = ii; // send it if not a null byte
 iPCXmiti++; // and point at next char *only* if not at null-end
} else if(bit_test(TXSTA,1)) // check to see if bits are all really shifted out
   {
    iPCXmitbsy = FALSE; //  we're done, flag to background
    bit_clear (TXSTA,5); // turn off port enable (TXEN)
    bit_clear (PIE1,4); //and interrupt enable (TXIE)
   }
} // no need to reset buffer index, this is done by transmit start code
// which sends first character, this sends the rest of them.

//////////////////////////////////**********
void TimeMachine()
{
  if (secflag)
  {
   secflag = FALSE; // we saw it
   if (++seconds >= 60) // belt and suspenders here, should never be 60
   { // minute boundary
    seconds = 0;  // start new minute
    if (++minutes >= 60)
    {
     minutes = 0;
     hours++; // will overflow at 65536 hours runtime.
    } // hour boundary end
   } // minute boundary end


// debug here once/second, app calls maybe
// else call apps after seconds loop -- fastest

   pwmduty = adres[0] >> 4; // 12 bits to 8 bits
   set_pwm1_duty (pwmduty); // start with a/d input for testing
   
     sprintf(PCXmitbuf,"\r\nRuntime: %Lu:%02u:%02u",hours,minutes, seconds);
     SendPC(PCXmitbuf); //iPCrcv[0]

  } // end if secflag


}

void PCMachine()
{
 switch (PCState)
 {
  case 0: // look for work
  if (ircvflag) PCState = 1; // Ok, handle it next time we are called
  break;
  case 1: // message available, do something about it
  // for now, just repeat it back, later, add another switch here to parse it
  strcpy(PCXmitbuf,PCrcv ); // copy sring to pc output buffer
  SendPC(PCXmitbuf); // not very clean -- doesn't check for busy yet  
// but hey, this is just debug for the moment
  ircvflag = FALSE; // we gitted it
   PCState = 0; // go back to look for work state
  break;
  
  
  default:  // should never get here
  printf("\r\nargh, PC machine borked");
  break;
 }
}


/***************************************************************************/
// Main.  This loops forever calling application state machines in turn
/////////////////////////////////////////////////////////////////////////////
void main()
{
 init();  // do hardware & memory init

  for(;;) // forever loop, call app state machines and track runtime
  {  
   restart_wdt(); // trying with watchdog timer on now.  
   // You get stupid and block in a wait loop, we'll reset your sorry butt.
   OUTPUT_TOGGLE (PIN_A6);  // Show we were here, pin 10 on 18lf2523
    
////////////////// fast app calls here  //////////////////////////////
// calls here get hit as fast as possible, once per roundy round
// return as quick as you can from each call so other apps get cycles.

   TimeMachine(); // track seconds flag, time since start
   PCMachine();  // handle rs232 input from PC host
  } // forever loop end
// can't get here unless a disaster
} // end of main

